@@ -3,8 +3,10 @@ package k8shandler
 import (
 	"context"
 	"fmt"
+	"reflect"
 	"time"
 
+	"github.com/openshift/elasticsearch-operator/pkg/elasticsearch"
 	"github.com/openshift/elasticsearch-operator/pkg/logger"
 	"github.com/openshift/elasticsearch-operator/pkg/utils"
 	"github.com/openshift/elasticsearch-operator/pkg/utils/comparators"
@@ -29,14 +31,16 @@ type statefulSetNode struct {
 	// prior hash for secret content
 	secretHash string
 
-	clusterName       string
-	clusterSize       int32
-	priorReplicaCount int32
+	clusterName string
+	clusterSize int32
+	//priorReplicaCount int32
 
 	client client.Client
+
+	esClient elasticsearch.Client
 }
 
-func (statefulSetNode *statefulSetNode) populateReference(nodeName string, node api.ElasticsearchNode, cluster *api.Elasticsearch, roleMap map[api.ElasticsearchNodeRole]bool, replicas int32, client client.Client) {
+func (statefulSetNode *statefulSetNode) populateReference(nodeName string, node api.ElasticsearchNode, cluster *api.Elasticsearch, roleMap map[api.ElasticsearchNodeRole]bool, replicas int32, client client.Client, esClient elasticsearch.Client) {
 
 	labels := newLabels(cluster.Name, nodeName, roleMap)
 
@@ -53,13 +57,13 @@ func (statefulSetNode *statefulSetNode) populateReference(nodeName string, node
 	}
 
 	partition := int32(0)
-
+	logConfig := getLogConfig(cluster.GetAnnotations())
 	statefulSet.Spec = apps.StatefulSetSpec{
 		Replicas: &replicas,
 		Selector: &metav1.LabelSelector{
 			MatchLabels: newLabelSelector(cluster.Name, nodeName, roleMap),
 		},
-		Template: newPodTemplateSpec(nodeName, cluster.Name, cluster.Namespace, node, cluster.Spec.Spec, labels, roleMap, client),
+		Template: newPodTemplateSpec(nodeName, cluster.Name, cluster.Namespace, node, cluster.Spec.Spec, labels, roleMap, client, logConfig),
 		UpdateStrategy: apps.StatefulSetUpdateStrategy{
 			Type: apps.RollingUpdateStatefulSetStrategyType,
 			RollingUpdate: &apps.RollingUpdateStatefulSetStrategy{
@@ -75,6 +79,7 @@ func (statefulSetNode *statefulSetNode) populateReference(nodeName string, node
 	statefulSetNode.clusterName = cluster.Name
 
 	statefulSetNode.client = client
+	statefulSetNode.esClient = esClient
 }
 
 func (current *statefulSetNode) updateReference(desired NodeTypeInterface) {
@@ -126,7 +131,7 @@ func (node *statefulSetNode) name() string {
 
 func (node *statefulSetNode) waitForNodeRejoinCluster() (error, bool) {
 	err := wait.Poll(time.Second*1, time.Second*60, func() (done bool, err error) {
-		clusterSize, getErr := GetClusterNodeCount(node.clusterName, node.self.Namespace, node.client)
+		clusterSize, getErr := node.esClient.GetClusterNodeCount()
 		if err != nil {
 			logrus.Warnf("Unable to get cluster size waiting for %v to rejoin cluster", node.name())
 			return false, getErr
@@ -140,7 +145,7 @@ func (node *statefulSetNode) waitForNodeRejoinCluster() (error, bool) {
 
 func (node *statefulSetNode) waitForNodeLeaveCluster() (error, bool) {
 	err := wait.Poll(time.Second*1, time.Second*60, func() (done bool, err error) {
-		clusterSize, getErr := GetClusterNodeCount(node.clusterName, node.self.Namespace, node.client)
+		clusterSize, getErr := node.esClient.GetClusterNodeCount()
 		if err != nil {
 			logrus.Warnf("Unable to get cluster size waiting for %v to leave cluster", node.name())
 			return false, getErr
@@ -246,12 +251,12 @@ func (node *statefulSetNode) isMissing() bool {
 func (node *statefulSetNode) rollingRestart(upgradeStatus *api.ElasticsearchNodeStatus) {
 
 	if upgradeStatus.UpgradeStatus.UnderUpgrade != v1.ConditionTrue {
-		if status, _ := GetClusterHealthStatus(node.clusterName, node.self.Namespace, node.client); !utils.Contains(desiredClusterStates, status) {
+		if status, _ := node.esClient.GetClusterHealthStatus(); !utils.Contains(desiredClusterStates, status) {
 			logrus.Infof("Waiting for cluster to be recovered before restarting %s: %s / %v", node.name(), status, desiredClusterStates)
 			return
 		}
 
-		size, err := GetClusterNodeCount(node.clusterName, node.self.Namespace, node.client)
+		size, err := node.esClient.GetClusterNodeCount()
 		if err != nil {
 			logrus.Warnf("Unable to get cluster size prior to restart for %v", node.name())
 			return
@@ -264,15 +269,15 @@ func (node *statefulSetNode) rollingRestart(upgradeStatus *api.ElasticsearchNode
 			return
 		}
 
-		node.setPartition(replicas)
+		if err := node.setPartition(replicas); err != nil {
+			logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+		}
 		upgradeStatus.UpgradeStatus.UnderUpgrade = v1.ConditionTrue
 	}
 
 	if upgradeStatus.UpgradeStatus.UpgradePhase == "" ||
 		upgradeStatus.UpgradeStatus.UpgradePhase == api.ControllerUpdated {
 
-		// nothing to do here -- just maintaing a framework structure
-
 		upgradeStatus.UpgradeStatus.UpgradePhase = api.NodeRestarting
 	}
 
@@ -281,7 +286,9 @@ func (node *statefulSetNode) rollingRestart(upgradeStatus *api.ElasticsearchNode
 		// if the node doesn't exist -- create it
 		// TODO: we can skip this logic after
 		if node.isMissing() {
-			node.create()
+			if err := node.create(); err != nil {
+				logrus.Warnf("unable to create a node. E: %s\r\n", err.Error())
+			}
 		}
 
 		ordinal, err := node.partition()
@@ -313,7 +320,9 @@ func (node *statefulSetNode) rollingRestart(upgradeStatus *api.ElasticsearchNode
 			}
 
 			// used for tracking in case of timeout
-			node.setPartition(index - 1)
+			if err := node.setPartition(index - 1); err != nil {
+				logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+			}
 		}
 
 		if err, _ := node.waitForNodeRejoinCluster(); err != nil {
@@ -342,13 +351,15 @@ func (node *statefulSetNode) fullClusterRestart(upgradeStatus *api.Elasticsearch
 			return
 		}
 
-		size, err := GetClusterNodeCount(node.clusterName, node.self.Namespace, node.client)
+		size, err := node.esClient.GetClusterNodeCount()
 		if err != nil {
 			logrus.Warnf("Unable to get cluster size prior to restart for %v", node.name())
 			return
 		}
 
-		node.setPartition(replicas)
+		if err := node.setPartition(replicas); err != nil {
+			logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+		}
 		node.clusterSize = size
 		upgradeStatus.UpgradeStatus.UnderUpgrade = v1.ConditionTrue
 	}
@@ -386,7 +397,9 @@ func (node *statefulSetNode) fullClusterRestart(upgradeStatus *api.Elasticsearch
 			}
 
 			// used for tracking in case of timeout
-			node.setPartition(index - 1)
+			if err := node.setPartition(index - 1); err != nil {
+				logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+			}
 		}
 
 		node.refreshHashes()
@@ -401,8 +414,8 @@ func (node *statefulSetNode) fullClusterRestart(upgradeStatus *api.Elasticsearch
 	}
 }
 
-func (node *statefulSetNode) delete() {
-	node.client.Delete(context.TODO(), &node.self)
+func (node *statefulSetNode) delete() error {
+	return node.client.Delete(context.TODO(), &node.self)
 }
 
 func (node *statefulSetNode) create() error {
@@ -445,12 +458,12 @@ func (node *statefulSetNode) executeUpdate() error {
 
 func (node *statefulSetNode) update(upgradeStatus *api.ElasticsearchNodeStatus) error {
 	if upgradeStatus.UpgradeStatus.UnderUpgrade != v1.ConditionTrue {
-		if status, _ := GetClusterHealthStatus(node.clusterName, node.self.Namespace, node.client); !utils.Contains(desiredClusterStates, status) {
+		if status, _ := node.esClient.GetClusterHealthStatus(); !utils.Contains(desiredClusterStates, status) {
 			logrus.Infof("Waiting for cluster to be recovered before restarting %s: %s / %v", node.name(), status, desiredClusterStates)
 			return fmt.Errorf("Waiting for cluster to be recovered before restarting %s: %s / %v", node.name(), status, desiredClusterStates)
 		}
 
-		size, err := GetClusterNodeCount(node.clusterName, node.self.Namespace, node.client)
+		size, err := node.esClient.GetClusterNodeCount()
 		if err != nil {
 			logrus.Warnf("Unable to get cluster size prior to restart for %v", node.name())
 		}
@@ -462,7 +475,9 @@ func (node *statefulSetNode) update(upgradeStatus *api.ElasticsearchNodeStatus)
 			return fmt.Errorf("Unable to get number of replicas prior to restart for %v", node.name())
 		}
 
-		node.setPartition(replicas)
+		if err := node.setPartition(replicas); err != nil {
+			logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+		}
 		upgradeStatus.UpgradeStatus.UnderUpgrade = v1.ConditionTrue
 	}
 
@@ -495,7 +510,9 @@ func (node *statefulSetNode) update(upgradeStatus *api.ElasticsearchNodeStatus)
 			}
 
 			// update partition to cause next pod to be updated
-			node.setPartition(index - 1)
+			if err := node.setPartition(index - 1); err != nil {
+				logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+			}
 
 			// wait for the node to leave the cluster
 			if err, _ := node.waitForNodeLeaveCluster(); err != nil {
@@ -551,7 +568,9 @@ func (node *statefulSetNode) scale() {
 		node.self.Spec.Replicas = desired.Spec.Replicas
 		logrus.Infof("Resource '%s' has different container replicas than desired", node.self.Name)
 
-		node.setReplicaCount(*node.self.Spec.Replicas)
+		if err := node.setReplicaCount(*node.self.Spec.Replicas); err != nil {
+			logrus.Warnf("unable to set replicate count. E: %s\r\n", err.Error())
+		}
 	}
 }
 
@@ -596,7 +615,11 @@ func (node *statefulSetNode) isChanged() bool {
 		if nodeContainer.Resources.Limits == nil {
 			nodeContainer.Resources.Limits = v1.ResourceList{}
 		}
-
+		if !reflect.DeepEqual(desiredContainer.Args, nodeContainer.Args) {
+			nodeContainer.Args = desiredContainer.Args
+			logger.Debugf("Container Args are different between current and desired for %s", nodeContainer.Name)
+			changed = true
+		}
 		// check that both exist
 
 		if nodeContainer.Image != desiredContainer.Image {
@@ -631,7 +654,13 @@ func (node *statefulSetNode) isChanged() bool {
 
 		if !comparators.EnvValueEqual(desiredContainer.Env, nodeContainer.Env) {
 			nodeContainer.Env = desiredContainer.Env
-			logger.Debugf("Container EnvVars are different between current and desired for %s", nodeContainer.Name)
+			logger.Debugf("Container EnvVars are different between current and desired for %s",
+				nodeContainer.Name)
+			changed = true
+		}
+		if !reflect.DeepEqual(desiredContainer.Ports, nodeContainer.Ports) {
+			nodeContainer.Ports = desiredContainer.Ports
+			logger.Debugf("Container Ports are different between current and desired for %s", nodeContainer.Name)
 			changed = true
 		}
 
@@ -662,3 +691,59 @@ func (node *statefulSetNode) progressUnshedulableNode(upgradeStatus *api.Elastic
 	}
 	return nil
 }
+
+func (node *statefulSetNode) progressNodeChanges(upgradeStatus *api.ElasticsearchNodeStatus) error {
+	if node.isChanged() {
+		replicas, err := node.replicaCount()
+		if err != nil {
+			logrus.Warnf("Unable to get number of replicas prior to restart for %v", node.name())
+			return fmt.Errorf("Unable to get number of replicas prior to restart for %v", node.name())
+		}
+
+		if err := node.setPartition(replicas); err != nil {
+			logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+		}
+
+		if err := node.executeUpdate(); err != nil {
+			return err
+		}
+
+		ordinal, err := node.partition()
+		if err != nil {
+			logrus.Infof("Unable to get node ordinal value: %v", err)
+			return err
+		}
+
+		// start partition at replicas and incrementally update it to 0
+		// making sure nodes rejoin between each one
+		for index := ordinal; index > 0; index-- {
+
+			// make sure we have all nodes in the cluster first -- always
+			if err, _ := node.waitForNodeRejoinCluster(); err != nil {
+				logrus.Infof("Timed out waiting for %v to rejoin cluster", node.name())
+				return fmt.Errorf("Timed out waiting for %v to rejoin cluster", node.name())
+			}
+
+			// update partition to cause next pod to be updated
+			if err := node.setPartition(index - 1); err != nil {
+				logrus.Warnf("unable to set partition. E: %s\r\n", err.Error())
+			}
+
+			// wait for the node to leave the cluster
+			if err, _ := node.waitForNodeLeaveCluster(); err != nil {
+				logrus.Infof("Timed out waiting for %v to leave the cluster", node.name())
+				return fmt.Errorf("Timed out waiting for %v to leave the cluster", node.name())
+			}
+		}
+
+		// this is here again because we need to make sure all nodes have rejoined
+		// before we move on and say we're done
+		if err, _ := node.waitForNodeRejoinCluster(); err != nil {
+			logrus.Infof("Timed out waiting for %v to rejoin cluster", node.name())
+			return fmt.Errorf("Timed out waiting for %v to rejoin cluster", node.name())
+		}
+
+		node.refreshHashes()
+	}
+	return nil
+}
