@@ -1,16 +1,53 @@
 package k8shandler
 
 import (
+	"context"
 	"fmt"
 	"strings"
 
-	"github.com/operator-framework/operator-sdk/pkg/sdk"
 	v1 "k8s.io/api/core/v1"
-
-	api "github.com/openshift/elasticsearch-operator/pkg/apis/elasticsearch/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/labels"
+	"sigs.k8s.io/controller-runtime/pkg/client"
+
+	api "github.com/openshift/elasticsearch-operator/pkg/apis/logging/v1"
 )
 
+const (
+	loglevelAnnotation          = "elasticsearch.openshift.io/loglevel"
+	serverLogAppenderAnnotation = "elasticsearch.openshift.io/develLogAppender"
+	serverLoglevelAnnotation    = "elasticsearch.openshift.io/esloglevel"
+)
+
+type LogConfig struct {
+	//LogLevel of the proxy and server security
+	LogLevel string
+	//ServerLogLevel of the remainder of Elasticsearch
+	ServerLoglevel string
+	//ServerAppender where to log messages
+	ServerAppender string
+}
+
+func getLogConfig(annotations map[string]string) LogConfig {
+	config := LogConfig{"info", "info", "console"}
+	if value, found := annotations[loglevelAnnotation]; found {
+		if strings.TrimSpace(value) != "" {
+			config.LogLevel = value
+		}
+	}
+	if value, found := annotations[serverLoglevelAnnotation]; found {
+		if strings.TrimSpace(value) != "" {
+			config.ServerLoglevel = value
+		}
+	}
+	if value, found := annotations[serverLogAppenderAnnotation]; found {
+		if strings.TrimSpace(value) != "" {
+			config.ServerAppender = value
+		}
+	}
+	return config
+}
+
 func selectorForES(nodeRole string, clusterName string) map[string]string {
 
 	return map[string]string{
@@ -19,13 +56,6 @@ func selectorForES(nodeRole string, clusterName string) map[string]string {
 	}
 }
 
-func labelsForESCluster(clusterName string) map[string]string {
-
-	return map[string]string{
-		"cluster-name": clusterName,
-	}
-}
-
 func appendDefaultLabel(clusterName string, labels map[string]string) map[string]string {
 	if _, ok := labels["cluster-name"]; ok {
 		return labels
@@ -66,13 +96,65 @@ func mergeSelectors(nodeSelectors, commonSelectors map[string]string) map[string
 	return commonSelectors
 }
 
-// getPodNames returns the pod names of the array of pods passed in
-func getPodNames(pods []v1.Pod) []string {
-	var podNames []string
-	for _, pod := range pods {
-		podNames = append(podNames, pod.Name)
+func areTolerationsSame(lhs, rhs []v1.Toleration) bool {
+
+	// if we are checking this as a part of pod spec comparison during a rollout we can't check this
+	// if we are comparing the deployment specs we can...
+	if len(lhs) != len(rhs) {
+		return false
 	}
-	return podNames
+
+	return containsSameTolerations(lhs, rhs)
+}
+
+// containsSameTolerations checks that the tolerations in lhs are all contained within rhs
+func containsSameTolerations(lhs, rhs []v1.Toleration) bool {
+
+	for _, lhsToleration := range lhs {
+		if !containsToleration(lhsToleration, rhs) {
+			return false
+		}
+	}
+
+	return true
+}
+
+func containsToleration(toleration v1.Toleration, tolerations []v1.Toleration) bool {
+	for _, t := range tolerations {
+		if isTolerationSame(t, toleration) {
+			return true
+		}
+	}
+
+	return false
+}
+
+func isTolerationSame(lhs, rhs v1.Toleration) bool {
+
+	tolerationSecondsBool := false
+	// check that both are either null or not null
+	if (lhs.TolerationSeconds == nil) == (rhs.TolerationSeconds == nil) {
+		if lhs.TolerationSeconds != nil {
+			// only compare values (attempt to dereference) if pointers aren't nil
+			tolerationSecondsBool = (*lhs.TolerationSeconds == *rhs.TolerationSeconds)
+		} else {
+			tolerationSecondsBool = true
+		}
+	}
+
+	return (lhs.Key == rhs.Key) &&
+		(lhs.Operator == rhs.Operator) &&
+		(lhs.Value == rhs.Value) &&
+		(lhs.Effect == rhs.Effect) &&
+		tolerationSecondsBool
+}
+
+func appendTolerations(nodeTolerations, commonTolerations []v1.Toleration) []v1.Toleration {
+	if commonTolerations == nil {
+		commonTolerations = []v1.Toleration{}
+	}
+
+	return append(commonTolerations, nodeTolerations...)
 }
 
 func getMasterCount(dpl *api.Elasticsearch) int32 {
@@ -93,21 +175,9 @@ func getDataCount(dpl *api.Elasticsearch) int32 {
 			dataCount = dataCount + node.NodeCount
 		}
 	}
-
 	return dataCount
 }
 
-func getClientCount(dpl *api.Elasticsearch) int32 {
-	clientCount := int32(0)
-	for _, node := range dpl.Spec.Nodes {
-		if isClientNode(node) {
-			clientCount = clientCount + node.NodeCount
-		}
-	}
-
-	return clientCount
-}
-
 func isValidMasterCount(dpl *api.Elasticsearch) bool {
 
 	if len(dpl.Spec.Nodes) == 0 {
@@ -144,47 +214,51 @@ func isValidRedundancyPolicy(dpl *api.Elasticsearch) bool {
 	return !(dataCount == 1 && dpl.Spec.RedundancyPolicy == api.SingleRedundancy)
 }
 
-func isValidConf(dpl *api.Elasticsearch) error {
+func (elasticsearchRequest *ElasticsearchRequest) isValidConf() error {
+
+	dpl := elasticsearchRequest.cluster
+	client := elasticsearchRequest.client
+
 	if !isValidMasterCount(dpl) {
-		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidMasterCountCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidMasterCountCondition, client); err != nil {
 			return err
 		}
 		return fmt.Errorf("Invalid master nodes count. Please ensure there are no more than %v total nodes with master roles", maxMasterCount)
 	} else {
-		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidMasterCountCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidMasterCountCondition, client); err != nil {
 			return err
 		}
 	}
 
 	if !isValidDataCount(dpl) {
-		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidDataCountCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidDataCountCondition, client); err != nil {
 			return err
 		}
 		return fmt.Errorf("No data nodes requested. Please ensure there is at least 1 node with data roles")
 	} else {
-		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidDataCountCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidDataCountCondition, client); err != nil {
 			return err
 		}
 	}
 
 	if !isValidRedundancyPolicy(dpl) {
-		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidReplicationCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionTrue, updateInvalidReplicationCondition, client); err != nil {
 			return err
 		}
 		return fmt.Errorf("Wrong RedundancyPolicy selected '%s'. Choose different RedundancyPolicy or add more nodes with data roles", dpl.Spec.RedundancyPolicy)
 	} else {
-		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidReplicationCondition); err != nil {
+		if err := updateConditionWithRetry(dpl, v1.ConditionFalse, updateInvalidReplicationCondition, client); err != nil {
 			return err
 		}
 	}
 
 	if ok, msg := hasValidUUIDs(dpl); !ok {
-		if err := updateInvalidUUIDChangeCondition(dpl, v1.ConditionTrue, msg); err != nil {
+		if err := updateInvalidUUIDChangeCondition(dpl, v1.ConditionTrue, msg, client); err != nil {
 			return err
 		}
 		return fmt.Errorf("Unsupported change to UUIDs made: %v", msg)
 	} else {
-		if err := updateInvalidUUIDChangeCondition(dpl, v1.ConditionFalse, ""); err != nil {
+		if err := updateInvalidUUIDChangeCondition(dpl, v1.ConditionFalse, "", client); err != nil {
 			return err
 		}
 	}
@@ -265,7 +339,7 @@ func sliceContainsString(slice []string, value string) bool {
 	return false
 }
 
-func DeletePod(podName, namespace string) error {
+func DeletePod(podName, namespace string, client client.Client) error {
 	pod := &v1.Pod{
 		TypeMeta: metav1.TypeMeta{
 			Kind:       "Pod",
@@ -277,25 +351,20 @@ func DeletePod(podName, namespace string) error {
 		},
 	}
 
-	err := sdk.Delete(pod)
+	err := client.Delete(context.TODO(), pod)
 
 	return err
 }
 
-func GetPodList(namespace string, selector string) (*v1.PodList, error) {
-	list := &v1.PodList{
-		TypeMeta: metav1.TypeMeta{
-			Kind:       "Pod",
-			APIVersion: v1.SchemeGroupVersion.String(),
-		},
-	}
+func GetPodList(namespace string, selector map[string]string, sdkClient client.Client) (*v1.PodList, error) {
+	list := &v1.PodList{}
 
-	err := sdk.List(
-		namespace,
+	labelSelector := labels.SelectorFromSet(selector)
+
+	err := sdkClient.List(
+		context.TODO(),
+		&client.ListOptions{Namespace: namespace, LabelSelector: labelSelector},
 		list,
-		sdk.WithListOptions(&metav1.ListOptions{
-			LabelSelector: selector,
-		}),
 	)
 
 	return list, err
